@@ -36,20 +36,26 @@ static DConfEngineServiceFunc dconf_engine_service_func;
 void
 dconf_engine_message_destroy (DConfEngineMessage *dcem)
 {
-  if (dcem->body)
-    {
-      g_variant_unref (dcem->body);
-      dcem->body = NULL;
-    }
+  gint i;
+
+  for (i = 0; dcem->parameters[i]; i++)
+    g_variant_unref (dcem->parameters[i]);
+  g_free (dcem->parameters);
 }
 
 void
 dconf_engine_message_copy (DConfEngineMessage *orig,
                            DConfEngineMessage *copy)
 {
+  gint i, n;
+
   *copy = *orig;
-  if (orig->body)
-    copy->body = g_variant_ref (orig->body);
+
+  for (n = 0; orig->parameters[n]; n++);
+  copy->parameters = g_new (GVariant *, n + 1);
+  for (i = 0; i < n; i++)
+    copy->parameters[i] = g_variant_ref (orig->parameters[i]);
+  copy->parameters[i] = NULL;
 }
 
 void
@@ -71,20 +77,25 @@ dconf_engine_get_session_dir (void)
       if (session_dir == NULL)
         {
           DConfEngineMessage dcem;
+          GVariant *parameters[2];
           GVariant *result;
 
-          dcem.bus_type = 'e';
-          dcem.destination = "ca.desrt.dconf";
+          dcem.bus_types = "e";
+          dcem.bus_name = "ca.desrt.dconf";
           dcem.object_path = "/ca/desrt/dconf/Writer";
-          dcem.interface = "org.freedesktop.DBus.Properties";
-          dcem.method = "Get";
+          dcem.interface_name = "org.freedesktop.DBus.Properties";
+          dcem.method_name = "Get";
           dcem.reply_type = G_VARIANT_TYPE ("(v)");
-          dcem.body = g_variant_new ("((ss))",
-                                     "ca.desrt.dconf.WriterInfo",
-                                     "ShmDirectory");
+          parameters[0] = g_variant_new ("(ss)",
+                                         "ca.desrt.dconf.WriterInfo",
+                                         "ShmDirectory");
+          parameters[1] = NULL;
+          dcem.parameters = parameters;
 
           result = dconf_engine_service_func (&dcem);
 
+          g_variant_unref (parameters[0]);
+
           if (result != NULL)
             {
               GVariant *str;
@@ -111,12 +122,15 @@ dconf_engine_get_session_dir (void)
 
 struct _DConfEngine
 {
+  GStaticMutex lock;
   guint64     state;
 
   guint8     *shm;
 
   GvdbTable **gvdbs;
+  GvdbTable **lock_tables;
   gchar     **object_paths;
+  gchar      *bus_types;
   gchar     **names;
   gint        n_dbs;
 };
@@ -144,7 +158,7 @@ dconf_engine_setup_user (DConfEngine *engine)
 
           if (fd >= 0)
             {
-              if (posix_fallocate (fd, 0, 1) == 0)
+              if (ftruncate (fd, 1) == 0)
                 {
                   engine->shm = mmap (NULL, 1, PROT_READ, MAP_SHARED, fd, 0);
 
@@ -217,6 +231,8 @@ dconf_engine_refresh_system (DConfEngine *engine)
           if (engine->gvdbs[i] == NULL)
             g_error ("Unable to open '%s', specified in dconf profile\n",
                      filename);
+          engine->lock_tables[i] = gvdb_table_get_table (engine->gvdbs[i],
+                                                         ".locks");
           g_free (filename);
           engine->state++;
         }
@@ -233,9 +249,16 @@ dconf_engine_refresh (DConfEngine *engine)
 guint64
 dconf_engine_get_state (DConfEngine *engine)
 {
-  dconf_engine_refresh (engine);
+  guint64 state;
 
-  return engine->state;
+  g_static_mutex_lock (&engine->lock);
+
+  dconf_engine_refresh (engine);
+  state = engine->state;
+
+  g_static_mutex_unlock (&engine->lock);
+
+  return state;
 }
 
 static gboolean
@@ -306,6 +329,7 @@ dconf_engine_new (const gchar *profile)
   gint i;
 
   engine = g_slice_new (DConfEngine);
+  g_static_mutex_init (&engine->lock);
   engine->shm = NULL;
 
   if (profile == NULL)
@@ -339,6 +363,9 @@ dconf_engine_new (const gchar *profile)
 
   engine->object_paths = g_new (gchar *, engine->n_dbs);
   engine->gvdbs = g_new0 (GvdbTable *, engine->n_dbs);
+  engine->lock_tables = g_new0 (GvdbTable *, engine->n_dbs);
+  engine->bus_types = g_strdup ("eyyyyyyyyyyyyy");
+  engine->state = 0;
 
   for (i = 0; i < engine->n_dbs; i++)
     if (engine->names[i])
@@ -367,6 +394,9 @@ dconf_engine_free (DConfEngine *engine)
 
       if (engine->gvdbs[i])
         gvdb_table_unref (engine->gvdbs[i]);
+
+      if (engine->lock_tables[i])
+        gvdb_table_unref (engine->lock_tables[i]);
     }
 
   if (engine->shm)
@@ -374,76 +404,119 @@ dconf_engine_free (DConfEngine *engine)
       munmap (engine->shm, 1);
     }
 
+  g_static_mutex_free (&engine->lock);
+
   g_free (engine->object_paths);
+  g_free (engine->bus_types);
   g_free (engine->names);
   g_free (engine->gvdbs);
+  g_free (engine->lock_tables);
 
   g_slice_free (DConfEngine, engine);
 }
 
+static GVariant *
+dconf_engine_read_internal (DConfEngine  *engine,
+                            const gchar  *key,
+                            gboolean      user,
+                            gboolean      system)
+{
+  GVariant *value = NULL;
+  gint lowest;
+  gint limit;
+  gint i;
+
+  g_static_mutex_lock (&engine->lock);
+
+  dconf_engine_refresh (engine);
+
+  /* Bound the search space depending on the databases that we are
+   * interested in.
+   */
+  limit = system ? engine->n_dbs : 1;
+  lowest = user ? 0 : 1;
+
+  /* We want i equal to the index of the highest database containing a
+   * lock, or i == lowest if there is no lock.  For that reason, we
+   * don't actually check the lowest database for a lock.  That makes
+   * sense, because even if it had a lock, it would not change our
+   * search policy (which would be to check the lowest one first).
+   *
+   * Note that we intentionally dishonour 'limit' here -- we want to
+   * ensure that values in the user database are always ignored when
+   * locks are present.
+   */
+  for (i = engine->n_dbs - 1; lowest < i; i--)
+    if (engine->lock_tables[i] != NULL &&
+        gvdb_table_has_value (engine->lock_tables[i], key))
+      break;
+
+  while (i < limit && value == NULL)
+    {
+      if (engine->gvdbs[i] != NULL)
+        value = gvdb_table_get_value (engine->gvdbs[i], key);
+      i++;
+    }
+
+  g_static_mutex_unlock (&engine->lock);
+
+  return value;
+}
+
 GVariant *
 dconf_engine_read (DConfEngine  *engine,
                    const gchar  *key)
 {
-  GVariant *value = NULL;
-  gint i;
-
-  dconf_engine_refresh (engine);
-
-  if (engine->gvdbs[0])
-    value = gvdb_table_get_value (engine->gvdbs[0], key);
-
-  for (i = 1; i < engine->n_dbs && value == NULL; i++)
-    value = gvdb_table_get_value (engine->gvdbs[i], key);
-
-  return value;
+  return dconf_engine_read_internal (engine, key, TRUE, TRUE);
 }
 
 GVariant *
 dconf_engine_read_default (DConfEngine  *engine,
                            const gchar  *key)
 {
-  GVariant *value = NULL;
-  gint i;
-
-  dconf_engine_refresh_system (engine);
-
-  for (i = 1; i < engine->n_dbs && value == NULL; i++)
-    value = gvdb_table_get_value (engine->gvdbs[i], key);
-
-  return value;
+  return dconf_engine_read_internal (engine, key, FALSE, TRUE);
 }
 
 GVariant *
 dconf_engine_read_no_default (DConfEngine  *engine,
                               const gchar  *key)
 {
-  GVariant *value = NULL;
-
-  dconf_engine_refresh_user (engine);
-
-  if (engine->gvdbs[0])
-    value = gvdb_table_get_value (engine->gvdbs[0], key);
-
-  return value;
+  return dconf_engine_read_internal (engine, key, TRUE, FALSE);
 }
 
 static void
 dconf_engine_make_match_rule (DConfEngine        *engine,
                               DConfEngineMessage *dcem,
-                              const gchar        *name)
+                              const gchar        *name,
+                              const gchar        *method_name)
 {
-  gchar *rule;
+  gint i;
 
-  rule = g_strdup_printf ("type='signal',interface='ca.desrt.dconf.Writer',"
-                          "path='%s',arg0path='%s'",
-                          engine->object_paths[0], name);
-  dcem->bus_type = 'e';
-  dcem->destination = "org.freedesktop.DBus";
+  dcem->bus_name = "org.freedesktop.DBus";
   dcem->object_path = "/org/freedesktop/DBus";
-  dcem->interface = "org.freedesktop.DBus";
-  dcem->body = g_variant_ref_sink (g_variant_new ("((s))", rule));
-  g_free (rule);
+  dcem->interface_name = "org.freedesktop.DBus";
+  dcem->method_name = method_name;
+
+  dcem->parameters = g_new (GVariant *, engine->n_dbs + 1);
+  for (i = 0; i < engine->n_dbs; i++)
+    {
+      gchar *rule;
+
+      rule = g_strdup_printf ("type='signal',"
+                              "interface='ca.desrt.dconf.Writer',"
+                              "path='%s',"
+                              "arg0path='%s'",
+                              engine->object_paths[i],
+                              name);
+      dcem->parameters[i] = g_variant_new ("(s)", rule);
+      g_variant_ref_sink (dcem->parameters[i]);
+      g_free (rule);
+    }
+  dcem->parameters[i] = NULL;
+
+  dcem->bus_types = engine->bus_types;
+  dcem->n_messages = engine->n_dbs;
+  dcem->reply_type = G_VARIANT_TYPE_UNIT;
 }
 
 void
@@ -451,8 +524,7 @@ dconf_engine_watch (DConfEngine        *engine,
                     const gchar        *name,
                     DConfEngineMessage *dcem)
 {
-  dconf_engine_make_match_rule (engine, dcem, name);
-  dcem->method = "AddMatch";
+  dconf_engine_make_match_rule (engine, dcem, name, "AddMatch");
 }
 
 void
@@ -460,15 +532,59 @@ dconf_engine_unwatch (DConfEngine        *engine,
                       const gchar        *name,
                       DConfEngineMessage *dcem)
 {
-  dconf_engine_make_match_rule (engine, dcem, name);
-  dcem->method = "RemoveMatch";
+  dconf_engine_make_match_rule (engine, dcem, name, "RemoveMatch");
 }
 
 gboolean
-dconf_engine_is_writable (DConfEngine         *engine,
-                          const gchar         *name)
+dconf_engine_is_writable (DConfEngine *engine,
+                          const gchar *name)
 {
-  return TRUE;
+  gboolean writable = TRUE;
+
+  /* Only check if we have at least one system database */
+  if (engine->n_dbs > 1)
+    {
+      gint i;
+
+      g_static_mutex_lock (&engine->lock);
+
+      dconf_engine_refresh_system (engine);
+
+      /* Don't check for locks in the user database (i == 0). */
+      for (i = engine->n_dbs - 1; 0 < i; i--)
+        if (engine->lock_tables[i] != NULL &&
+            gvdb_table_has_value (engine->lock_tables[i], name))
+          {
+            writable = FALSE;
+            break;
+          }
+
+      g_static_mutex_unlock (&engine->lock);
+    }
+
+  return writable;
+}
+
+/* be conservative and fast:  false negatives are OK */
+static gboolean
+is_dbusable (GVariant *value)
+{
+  const gchar *type;
+
+  type = g_variant_get_type_string (value);
+
+  /* maybe definitely won't work.
+   * variant?  too lazy to check inside...
+   */
+  if (strchr (type, 'v') || strchr (type, 'm'))
+    return FALSE;
+
+  /* XXX: we could also check for '{}' not inside an array...
+   * but i'm not sure we want to support that anyway.
+   */
+
+  /* this will avoid any too-deeply-nested limits */
+  return strlen (type) < 32;
 }
 
 static GVariant *
@@ -479,7 +595,28 @@ fake_maybe (GVariant *value)
   g_variant_builder_init (&builder, G_VARIANT_TYPE ("av"));
 
   if (value != NULL)
-    g_variant_builder_add (&builder, "v", value);
+    {
+      if (is_dbusable (value))
+        g_variant_builder_add (&builder, "v", value);
+
+      else
+        {
+          GVariant *variant;
+          GVariant *ay;
+
+          variant = g_variant_new_variant (value);
+          ay = g_variant_new_from_data (G_VARIANT_TYPE_BYTESTRING,
+                                        g_variant_get_data (variant),
+                                        g_variant_get_size (variant),
+                                        TRUE,
+                                        (GDestroyNotify) g_variant_unref,
+                                        variant);
+          g_variant_builder_add (&builder, "v", ay);
+
+          g_variant_builder_add (&builder, "v",
+                                 g_variant_new_string ("serialised GVariant"));
+        }
+    }
 
   return g_variant_builder_end (&builder);
 }
@@ -487,23 +624,27 @@ fake_maybe (GVariant *value)
 static void
 dconf_engine_dcem (DConfEngine        *engine,
                    DConfEngineMessage *dcem,
-                   const gchar        *method,
+                   const gchar        *method_name,
                    const gchar        *format_string,
                    ...)
 {
   va_list ap;
 
-  dcem->bus_type = 'e';
-  dcem->destination = "ca.desrt.dconf";
+  dcem->bus_name = "ca.desrt.dconf";
   dcem->object_path = engine->object_paths[0];
-  dcem->interface = "ca.desrt.dconf.Writer";
-  dcem->reply_type = G_VARIANT_TYPE ("(s)");
-  dcem->method = method;
-  dcem->tagged = TRUE;
+  dcem->interface_name = "ca.desrt.dconf.Writer";
+  dcem->method_name = method_name;
+  dcem->parameters = g_new (GVariant *, 2);
+  dcem->n_messages = 1;
 
   va_start (ap, format_string);
-  dcem->body = g_variant_ref_sink (g_variant_new_variant (g_variant_new_va (format_string, NULL, &ap)));
+  dcem->parameters[0] = g_variant_new_va (format_string, NULL, &ap);
+  g_variant_ref_sink (dcem->parameters[0]);
+  dcem->parameters[1] = NULL;
   va_end (ap);
+
+  dcem->bus_types = engine->bus_types;
+  dcem->reply_type = G_VARIANT_TYPE ("(s)");
 }
 
 gboolean
@@ -562,6 +703,8 @@ dconf_engine_list (DConfEngine    *engine,
   /* not yet supported */
   g_assert (resets == NULL);
 
+  g_static_mutex_lock (&engine->lock);
+
   dconf_engine_refresh (engine);
 
   if (engine->gvdbs[0])
@@ -575,6 +718,8 @@ dconf_engine_list (DConfEngine    *engine,
   if (length)
     *length = g_strv_length (list);
 
+  g_static_mutex_unlock (&engine->lock);
+
   return list;
 }
 
@@ -610,6 +755,24 @@ dconf_engine_decode_notify (DConfEngine   *engine,
   return TRUE;
 }
 
+gboolean
+dconf_engine_decode_writability_notify (const gchar **path,
+                                        const gchar  *iface,
+                                        const gchar  *method,
+                                        GVariant     *body)
+{
+  if (strcmp (iface, "ca.desrt.dconf.Writer") ||
+      strcmp (method, "WritabilityNotify"))
+    return FALSE;
+
+  if (!g_variant_is_of_type (body, G_VARIANT_TYPE ("(s)")))
+    return FALSE;
+
+  g_variant_get_child (body, 0, "&s", path);
+
+  return TRUE;
+}
+
 gboolean
 dconf_engine_interpret_reply (DConfEngineMessage  *dcem,
                               const gchar         *sender,
